# Generated by Django 3.1.7 on 2021-10-12 11:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("permissions", "0001_initial"),
        ("flow", "0003_auto_20211012_1134"),
    ]

    operations = [
        migrations.AddField(
            model_name="collection",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="collection",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="data",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="data",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="descriptorschema",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="descriptorschema",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="entity",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entity",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="process",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="process",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="relation",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="relation",
                to="permissions.permissiongroup",
            ),
        ),
    ]
